/******************************************************************************
//(C) Copyright 2009 Marvell International Ltd.
//All Rights Reserved
******************************************************************************/

#ifndef _MRVL_LIST_H_
#define _MRVL_LIST_H_

#define MRVL_INLINE __inline

struct list_head {
    struct list_head *next, *prev;
};

#define INIT_LIST_HEAD(ptr) do { (ptr)->next = (ptr); (ptr)->prev = (ptr); } while (0)

static MRVL_INLINE void __mrvl_list_add(struct list_head * _new, struct list_head * prev, struct list_head * next)
{
    next->prev = _new;
    _new->next = next;
    _new->prev = prev;
    prev->next = _new;
}

static MRVL_INLINE void mrvl_list_add(struct list_head *_new, struct list_head *head)
{
    __mrvl_list_add(_new, head, head->next);
}

static MRVL_INLINE void list_add_tail(struct list_head *_new, struct list_head *head)
{
    __mrvl_list_add(_new, head->prev, head);
}

static MRVL_INLINE void __mrvl_list_del(struct list_head * prev, struct list_head * next)
{
    next->prev = prev;
    prev->next = next;
}

static MRVL_INLINE void mrvl_list_del(struct list_head *entry)
{
    __mrvl_list_del(entry->prev, entry->next);
}

static MRVL_INLINE void list_del_init(struct list_head *entry)
{
    __mrvl_list_del(entry->prev, entry->next);
    INIT_LIST_HEAD(entry); 
}

static MRVL_INLINE int list_empty(struct list_head *head)
{
    return head->next == head;
}

static MRVL_INLINE void __list_splice(struct list_head *list,
                 struct list_head *head)
{
    struct list_head *first = list->next;
    struct list_head *last = list->prev;
    struct list_head *at = head->next;

    first->prev = head;
    head->next = first;

    last->next = at;
    at->prev = last;
}

static MRVL_INLINE void list_splice(struct list_head *list, struct list_head *head)
{
    if (!list_empty(list))
        __list_splice(list, head);
}

static MRVL_INLINE void list_splice_init(struct list_head *list, struct list_head *head)
{
    if (!list_empty(list)) {
        __list_splice(list, head);
        INIT_LIST_HEAD(list);
    }
}

#define list_entry(ptr, type, member) \
    ((type *)((char *)(ptr)-(unsigned long)(&((type *)0)->member)))

#define list_for_each(pos, head) \
    for (pos = (head)->next; pos != (head); pos = pos->next)

#define list_reverse_for_each(pos, head) \
    for (pos = (head)->prev; pos != (head); pos = pos->prev)

#define get_first_item(attached, type, member) \
    ((type *)((char *)((attached)->next)-(unsigned long)(&((type *)0)->member)))

#endif



